#ifndef METADATAHOLDER_H_
#define METADATAHOLDER_H_

#include <string>
#include <map>
#include <iostream>
#include "trace.h"
#include "exceptionBase.h"

namespace iopc {

class Metadata {
public:
	enum MetadataType {
		INT,
		STRING,
		BOOL,
		POINTER,
		NOTHING
	};

	Metadata() {
		type = NOTHING;
		intValue = 0;
		pointerValue = NULL;
		boolValue = false;
	}
	Metadata(int value) {
		setIntValue(value);
	}
	Metadata(const std::string& value) {
		setStringValue(value);
	}
	Metadata(const char* value) {
		setStringValue(value);
	}
	Metadata(bool value) {
		setBoolValue(value);
	}
	Metadata(void* value) {
		setPointerValue(value);
	}

	inline void setIntValue(int value) {
		type = INT;
		intValue = value;
	}
	inline void setStringValue(const std::string& value) {
		type = STRING;
		stringValue = value;
	}
	/*inline void setStringValue(const char* value) {
		type = STRING;
		stringValue = value;
	}*/
	inline void setBoolValue(bool value) {
		type = BOOL;
		boolValue = value;
	}
	inline void setPointerValue(void* value) {
		type = POINTER;
		pointerValue = value;
	}

	inline int getIntValue() const {
		if (!type == INT) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return intValue;
	}
	inline std::string getStringValue() const {
		if (!type == STRING) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return stringValue;
	}
	inline bool getBoolValue() const {
		if (!type == BOOL) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return boolValue;
	}
	inline void* getPointerValue() const {
		if (!type == POINTER) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Wrong type of metadata"));
		}
		return pointerValue;
	}

	operator const bool() const {
		return getBoolValue();
	}
	operator const int() const {
		return getIntValue();
	}
	operator const std::string() const {
		return getStringValue();
	}
	/*operator const char*() const {
		return getStringValue().c_str();
	}*/
	operator const void*() const {
		return getPointerValue();
	}

	inline bool isEmpty() const {
		return type == NOTHING;
	}

	// only the NOTHING type can be compared at the moment...
	bool operator==(const Metadata& src) {
		return (src.type == NOTHING) && (this->type == NOTHING);
	}

	std::string toString() const;

	friend std::ostream& operator<<(std::ostream& out, const Metadata& m);

private:
	int intValue;
	bool boolValue;
	void* pointerValue;
	std::string stringValue;
	MetadataType type;
};

std::ostream& operator<<(std::ostream& out, const Metadata& m);


class MetadataHolder {
private:
	/**
	 * Map containing metadata
	 */
	mutable std::map<std::string, Metadata> meta;
	const MetadataHolder* secondaryMeta;
public:
	MetadataHolder();

	static MetadataHolder& getDefaultsMeta();

	static inline bool hasDefaultMeta(const std::string& key) {
		return getDefaultsMeta().hasMeta(key);
	}

	virtual const Metadata& getMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;
	void setMeta(const std::string& key, const Metadata& value) const;
	void setSecondaryMetadataLocation(const MetadataHolder* secondaryMeta);
	virtual bool hasMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;
	// hasMeta && getBoolValue == true
	inline bool hasFeature(const std::string& key) const {
		return hasMeta(key) && getMeta(key).getBoolValue();
	}
	inline void setFeature(const std::string& key) const {
		meta[key].setBoolValue(true);
	}
	inline bool hasSecondaryMetadataLocation() const {
		return secondaryMeta != NULL;
	}

	/**
	 * Simple access to the metadata; suitable for setting values.
	 *
	 * This operator doesn't access the default values and secondary metadata location
	 * when looking up the metadata value!
	 * May return values with Metadata.type == NOTHING.
	 */
	inline Metadata& operator[](const std::string& key) const {
		return meta[key];
	}

	static MetadataHolder emptyMeta;

	virtual ~MetadataHolder() {};

	inline bool operator!=(const MetadataHolder& src) const { return this != &src; }
	inline bool operator==(const MetadataHolder& src) const { return this == &src; }
	friend std::ostream& operator <<(std::ostream& out, const MetadataHolder& m);
	std::string toString() const;
};
std::ostream& operator<<(std::ostream& out, const MetadataHolder& m);


class AdditionalMetadataHolder : public MetadataHolder
{
private:
	const MetadataHolder& primary;
	bool invert;
public:
	AdditionalMetadataHolder(const MetadataHolder& primary, bool invert = false) : primary(primary), invert(invert) {};
	virtual const Metadata& getMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;
	virtual bool hasMeta(const std::string& key, bool searchDefaults = true, bool searchSecondary = true) const;
};

}
#endif /*METADATAHOLDER_H_*/
