#include <pthread.h>
#include <errno.h>
#include "rwxLock.h"
#include "trace.h"

namespace iopc {

RWXLock::RWXLock()
	throw (IopcNoMemoryException, IopcRsrcLackException)
: wlocks(0), isXLock(false), mutex(), wlocker(0)
{
	switch (pthread_rwlock_init(&rwlock, NULL)) {
	case 0:
		return;
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EAGAIN:
	case EPERM:
		IOPC_LOG_THROW(IopcRsrcLackException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

RWXLock::~RWXLock()
	throw ()
{
	IOPC_ASSERT(wlocks == 0 && !isXLock);
	int res = pthread_rwlock_destroy(&rwlock); IOPC_ASSERT(res == 0);
}

void RWXLock::rlock()
	throw (IopcRsrcLackException, IopcDeadlockException, IopcNoMemoryException)
{
	mutex.lock();
	if ((wlocks > 0 || isXLock) && pthread_equal(pthread_self(), wlocker)) {
		mutex.unlock();
		IOPC_LOG_THROW(IopcDeadlockException());
	}
	mutex.unlock();
	switch (pthread_rwlock_rdlock(&rwlock)) {
	case 0:
		IOPC_ASSERT(wlocks == 0 && !isXLock);
		return;
	break;
	case EAGAIN:
		IOPC_LOG_THROW(IopcRsrcLackException());
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	break;
	default:
		IOPC_UNREACHABLE;
	}
}

void RWXLock::wlock()
	throw (IopcDeadlockException, IopcNoMemoryException)
{
	mutex.lock();
	if (pthread_equal(pthread_self(), wlocker)) {
		if (isXLock) {
			mutex.unlock();
			IOPC_LOG_THROW(IopcDeadlockException());
		} else if (wlocks > 0) {
			wlocks++;
			mutex.unlock();
			return;
		}
	}
	mutex.unlock();
	switch (pthread_rwlock_wrlock(&rwlock)) {
	case 0:
		mutex.lock();
		IOPC_ASSERT(wlocks == 0 && !isXLock);
		wlocks = 1;
		wlocker = pthread_self();
		mutex.unlock();
		return;
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void RWXLock::xlock()
	throw (IopcDeadlockException, IopcNoMemoryException)
{
	mutex.lock();
	if ((wlocks > 0 || isXLock) && pthread_equal(pthread_self(), wlocker)) {
		mutex.unlock();
		IOPC_LOG_THROW(IopcDeadlockException());
	}
	mutex.unlock();
	switch (pthread_rwlock_wrlock(&rwlock)) {
	case 0:
		mutex.lock();
		IOPC_ASSERT(wlocks == 0 && !isXLock);
		isXLock = true;
		wlocker = pthread_self();
		mutex.unlock();
		return;
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

bool RWXLock::tryRLock()
	throw (IopcRsrcLackException, IopcDeadlockException)
{
	mutex.lock();
	if ((wlocks > 0 || isXLock) && pthread_equal(pthread_self(), wlocker)) {
		mutex.unlock();
		return false;
	}
	mutex.unlock();
	switch (pthread_rwlock_tryrdlock(&rwlock)) {
	case 0:
		IOPC_ASSERT(wlocks == 0 && !isXLock);
		return true;
	break;
	case EBUSY:
		return false;
	break;
	case EAGAIN:
		IOPC_LOG_THROW(IopcRsrcLackException());
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	break;
	default:
		IOPC_UNREACHABLE;
	}
}

bool RWXLock::tryWLock()
	throw (IopcRsrcLackException, IopcDeadlockException)
{
	mutex.lock();
	if (pthread_equal(pthread_self(), wlocker)) {
		if (isXLock) {
			mutex.unlock();
			return false;
		} else if (wlocks > 0) {
			wlocks++;
			mutex.unlock();
			return true;
		}
	}
	mutex.unlock();
	switch (pthread_rwlock_trywrlock(&rwlock)) {
	case 0:
		mutex.lock();
		IOPC_ASSERT(wlocks == 0 && !isXLock);
		wlocks = 1;
		wlocker = pthread_self();
		mutex.unlock();
		return true;
	break;
	case EBUSY:
		return false;
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

bool RWXLock::tryXLock()
	throw (IopcDeadlockException)
{
	mutex.lock();
	if ((wlocks > 0 || isXLock) && pthread_equal(pthread_self(), wlocker)) {
		mutex.unlock();
		return false;
	}
	mutex.unlock();
	switch (pthread_rwlock_trywrlock(&rwlock)) {
	case 0:
		mutex.lock();
		IOPC_ASSERT(wlocks == 0 && !isXLock);
		isXLock = true;
		wlocker = pthread_self();
		mutex.unlock();
		return true;
	break;
	case EBUSY:
		return false;
	break;
	case ENOMEM:
		IOPC_LOG_THROW(IopcNoMemoryException());
	break;
	case EDEADLK:
		IOPC_LOG_THROW(IopcDeadlockException());
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void RWXLock::unlock()
	throw (IopcIncorrectUsageException)
{
	mutex.lock();
	//int res = 0;
	if (pthread_equal(pthread_self(), wlocker)) {
		switch (wlocks) {
		case 0:
			if (!isXLock) {
				break;
			}
			isXLock = false;
		break;
		case 1:
			IOPC_ASSERT(!isXLock);
			wlocks = 0;
		break;
		default:
			wlocks--;
			mutex.unlock();
			return;
		break;
		}
		IOPC_ASSERT(wlocks == 0 && !isXLock);
	}
	mutex.unlock();
	switch (pthread_rwlock_unlock(&rwlock)) {
	case 0:		/* not locked */
		return;
	break;
	case EPERM:	/* other thread has lock for reading or writing */
		IOPC_LOG_THROW(IopcIncorrectUsageException());
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

}/*namespace polite*/
