/**
* @file defs.h
*
* @author Michal Kopecky  POLiTE library
* @author Jan Hadrava
* @author Petr Cermak
*/

#ifndef IOPCTRACE_H
#define IOPCTRACE_H

#include <fstream>
#include <ostream>
#include <string>
#include "defs.h"
#include "mutex.h"
#include <cstdlib>

// ----- Comment or uncomment logging option
#ifdef ECLIPSE_DEBUG
#define IOPC_DEBUG
#define IOPC_TRACE
#endif

namespace iopc
{

class DLL_EXTERNAL LogWriter
{
public:
	enum MsgType {NONE = 0, INFO_SQL = 1, INFO = 2, WARNING = 4, ERROR = 8, TRACE = 16, DEBUG = 32};

	static const std::string defaultFileName;
	static const int defaultTraceMask;
	static const int defaultTraceMaskTTY;

	int traceMask;
	int traceMaskTTY;

	LogWriter();
	virtual ~LogWriter();

	void logMsgInst(MsgType mask, const std::string& message);
	void logLineInst(MsgType mask, std::string file, unsigned int line, std::string function, const std::string& message);

	void logOpenInst(const std::string& fileName = defaultFileName, bool append = true);
	void logCloseInst();
	void logFlushInst();

	std::string getMsgTypeStr(MsgType msgType);

	std::string fileName;


	static void initInstance();
	static void terminateInstance();

	static LogWriter& getInstance();

	static inline void logMsg(MsgType mask, const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsgInst(mask, message);
	};
	static inline void logLine(MsgType mask, std::string file, unsigned int line, std::string function, const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logLineInst(mask, file, line, function, message);
	};
	static inline void logOpen(const std::string& fileName = defaultFileName, bool append = true) {
		if (!isInitialized()) return;
		getInstance().logOpenInst(fileName, append);
	};
	static inline void logClose() {
		if (!isInitialized()) return;
		getInstance().logFlushInst(); getInstance().logCloseInst();
	};
	static inline void logInfo(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(INFO, message);
	};
	static inline void logWarning(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(WARNING, message);
	};
	static inline void logError(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(ERROR, message);
	};
	static inline void logSql(const std::string& message) {
		if (!isInitialized()) return;
		getInstance().logMsg(INFO_SQL, message);
	};

	static inline bool isInitialized() { return instance != NULL; }
private:
	static Mutex logLock;
	static LogWriter* instance;

	std::ofstream logStream;
	//int va_log( MsgType mask, const char *format, va_list argp );
};

//extern LogWriter IopcLogWriter;


#ifndef __FUNCTION__
#ifdef __func__
#define __FUNCTION__ __func__
#endif
#else
#define __FUNCTION__ ""
#endif


// ----- DEBUG
#ifndef IOPC_DEBUG
#define IOPC_DEBUGLOG(msg) ;
#else // #ifndef IOPC_DEBUG
#define IOPC_DEBUGLOG(msg) LogWriter::logLine(LogWriter::DEBUG, __FILE__, __LINE__, __FUNCTION__, msg);
#endif // #ifndef IOPC_DEBUG

// ----- TRACE
#ifdef IOPC_TRACE
#define IOPC_FUNC_INVOKED						\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, "Function invoked");
#define IOPC_FUNC_END(retval)								\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Function ends with return value: " + #retval)
#define IOPC_FUNC_ENDED								\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Function ended")
#define IOPC_LOG_THROW(exc)								\
	do {									\
	const std::exception& e = exc;	\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Throwing exception: " + #exc + " what: " + e.what());					\
	throw exc;							\
	} while (0)
#define IOPC_LOG_RETHROW(exc)	\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Rethrowing exception: " + #exc);	\
	throw ;
#define IOPC_LOG_CATCH(exc)							\
	catch (exc) {							\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, (std::string) "Exception caught: " + #exc);
#define IOPC_LOG_CAUGHT							\
	LogWriter::logLine(LogWriter::TRACE, __FILE__, __LINE__, __FUNCTION__, "Exception solved");		\
	}
#else // #ifdef IOPC_TRACE
#define IOPC_LOG_THROW(exc)								\
	do {									\
	throw exc;							\
	} while (0)
#define IOPC_FUNC_INVOKED ;
#define IOPC_FUNC_END(retval) ;
#define IOPC_FUNC_ENDED ;
#define IOPC_LOG_RETHROW(exc)	\
	throw;
#define IOPC_LOG_CATCH(exc)	\
catch (exc) {
#define IOPC_LOG_CAUGHT	\
	}
#endif	// #ifdef IOPC_TRACE



#define IOPC_LOG_CATCH_NEXT(exc)						\
	IOPC_LOG_CAUGHT							\
	IOPC_LOG_CATCH(exc)


// ----- Common macros
#define IOPC_ASSERT(cond)									\
	do {											\
	if (!(cond)) {										\
	LogWriter::logLine(LogWriter::ERROR, __FILE__, __LINE__, __FUNCTION__, (std::string) "ASSERT failed: " + #cond);				\
	exit(-1);									\
	}											\
	} while (0)
#define IOPC_UNREACHABLE									\
	do {											\
	LogWriter::logLine(LogWriter::ERROR, __FILE__, __LINE__, __FUNCTION__, "Unreachable code reached");				\
	exit(-1);									\
	} while (0)



} /*namespace polite*/

#endif //__L_TRACE_H__
