/**
 * @file connection.h
 * 
 * @author Michal Kopecky
 * @author Josef Troch Josef Troch (<a href="mailto:josef.troch@email.cz">josef.troch@email.cz</a>)
 * @author Jan Hadrava
 * @author Petr Cermak
 */

#ifndef BASICCONNECTION_H
#define BASICCONNECTION_H

#include "iopccommon/lock.h"
#include "iopccommon/defsPlatform.h"
#include "database.h"
#include "cursor.h"
#include "connectionImpl.h"
#include "connection.h"
#include "basicDatabase.h"
#include <string>
#include <set>

namespace iopc
{

class Database;
class Cursor;

/**
 * @brief Represents a connection to a SQL database.
 * 
 * This class encapsulates one connection to a database and
 * allows user to control transaction processing.
 * One of its purposes is to manage cursors that are created by executing
 * SQL via @a sql(const std::string& name) or getCursor(const std::string& name) methods.
 * 
 * @author Michal Kopecky
 * @author Petr Cermak
 */
class DLL_EXTERNAL BasicConnection : public Connection
{
	
public:
	typedef std::set<Cursor*> CursorSet;
	typedef CursorSet::iterator CursorSetIterator;
	
private:
	BasicDatabase& database;		//the database this connection is
	bool autoCommit;
	bool opened;
	
	static CursorSet allCursors;
	CursorSet cursors;
	void addCursor(Cursor* cur);
	void removeCursor(Cursor* cur);
	
	ConnectionImpl* implementation;
	
	Lock lock;
	static Lock allLock;
	
public:

	/** Sets object cache associated with the connection. */
	//void SetCache(Cache &cache)	throw ();
	/** Unsets object cache associated with the connection. */
	//void SetCache()	throw ();
	/** Temporarily blocks cache usage. */
	//void StopCache();
	/** Unblocks cache usage. */
	//void RestartCache();
	
	//BasicConnection();
	virtual ~BasicConnection();
	
	virtual Cursor* sql(const std::string& name);
	
	//virtual Cursor* getCursor(const std::string& name);
	virtual void returnCursor(Cursor* cur);
	
	virtual void closeAllCursors();
	
	virtual void sqlNonQuery(const std::string& name);
	
	/** Commits current transaction. Starts new one. */
	virtual void commit();
	/** Rollbacks current transaction.*/
	virtual void rollback();
	/** Creates a savepoint. */
	virtual void savepoint(const std::string& name);
	/** Rollbacks to savepoint. */
	virtual void rollbackToSavepoint(const std::string& name);
	/** Opens connection */
	virtual void open();
	/** Closes connection */
	virtual void close();
	/** Commits last transaction and disconnects from DB. */
	virtual void disconnect();
	/** Rollbacks current trans
	 * action and disconnects from DB. */
	virtual void abort();
	/** Turns on/off autocommit. */
	virtual void setAutoCommit(bool ac);
	/** Returns state of this connection */
	virtual bool isOpened();
	
	/** Passes an SQL command to DB. */
	virtual Connection& operator << (
	    const std::string& command
	);
	/** Passes an SQL command to DB. */
/*	virtual Connection& operator << (
	    const Cmd& command
	);*/
	
	//Tests, if the Database connections are the same
	virtual bool operator== (const Connection& dbConn) const;
	//Tests, if the Database connections are not the same
	virtual bool operator!= (const Connection& dbConn) const;

	/*CacheLock &MakePersistent(const PersIdentification &persId, Object &objectPtr);
	CacheLock &GetCacheLock(const PersIdentification &persId, bool rdonly = false);
	void DbDelete(const PersIdentification &persId);
	void CacheDelete(const PersIdentification &persId);*/
	
	inline Database& getDatabase() { return database; }
	
	inline Driver& getDriver() { return database.getDriver(); }
	
	virtual inline bool isAutoCommit() { return autoCommit; }
	
	inline ConnectionImpl* getImplementation() { return implementation; }
	
	BasicConnection(ConnectionImpl* implementation, BasicDatabase& database); /*, Cache *cache*/
	
	Connection* getBasicConnection();

/*
	// generates new OID and Serial Number for new object
	virtual bool doGetNextOID(long int& nxtoid, long int& nxtsn) = 0;
	// generates new Serial Number for object
	virtual bool doGetNextSN(long int curoid, long int& nxtsn) = 0;*/
	
};


}

#endif

