/**
 * @file driver.h
 *
 * @author Petr Cermak
 */

#ifndef DRIVER_H_
#define DRIVER_H_

#include "databaseImpl.h"
#include "database.h"
#include "iopccommon/mutex.h"
#include "iopccommon/metadataHolder.h"

#include <string>
#include <set>
#include <map>

namespace iopc {

class DriverExtension;

/**
 * @brief The driver class is an entry point to a database driver in the IOPC library.
 *
 * Drivers are managed by the @a DriverManager class.
 * A Driver is always implemented as a singleton. Its functionality is
 * to setup and configure underlying libraries and to provide acces to
 * instance of the @a Database class.
 *
 * This class is thread safe and can be used by more threads concurrently.
 *
 * @author Petr Cermak
 */

class Driver : public MetadataHolder {
	public:

		/**
		 * @brief Returns name of this driver
		 *
		 * @returns Name of this driver
		 */
		const std::string& getName() { return name; };

		/**
		 * @brief Returns an 'Database' object representing a physical database identified by the dbName in the DBMS.
		 *
		 * Just calling getDatabase doesn't create connection to the db, but some drivers
		 * may do platform-dependent resource allocation and/or environment preparation here.
		 *
		 * @param dbName Name of the physical database
		 * @returns Pointer to the database object. The pointer is owned by the caller and caller is responsible for freeing
		 * allocated resources by calling the @a Driver::returnDatabase(Database* db) when the database is no more needed.
		 * The returned object is a super-typed @a BasicDatabase instance.
		 *
		 * @see Database
		 * @see BasicDatabase
		 */
		Database* getDatabase(const std::string& dbName);

		/**
		 * @brief Deallocates the database object and all related resources.
		 *
		 * This method accepts also decorated instances returned by the @a Driver::getDatabase(const std::string& dbName) method.
		 * Example:
		 * <code>
		 * Database* db = new CachedDatabase(driver.getDatabase(""), cache);
		 * ...
		 * driver.returnDatabase(db);  // deletes CachedDatabase instance
		 *
		 * </code>
		 *
		 * Closes all connections to the db database and destroys the Database instance with all decorators.
		 *
		 * @param db Database to be deallocated.
		 */
		void returnDatabase(Database* db);

		/**
		 * @brief Runs the driver initialization block.
		 *
		 * This method is called automatically from inside the Iopc initialization routine (@a Iopc::init())
		 * Calls the @a doInit() method which must be implemented in the database driver.
		 *
		 * @see DriverManager::initDrivers()
		 */
		void init(std::map<std::string, DriverExtension*>& extensions);

		/**
		 * @brief Runs the driver termination block.
		 *
		 * This method is called automatically from inside the Iopc termination routine (@a Iopc::shutdown())
		 * Calls the @a doShutdown() method which must be implemented in the database driver.
		 *
		 * @see DriverManager::terminateDrivers()
		 */
		void shutdown();

		// features
		/**
		 * @brief Registers a new driver feature with implementation.
		 *
		 * Called usually from driver extension's initialization routine (@a DriverExtension::doInit(Driver& driver))
		 *
		 * @param feature Pointer to a feature that this driver will support.
		 * The pointer points directly to the implementation of the feature.
		 */
		template <typename T>
		void registerFeature(T& feature) {
			(*this)[std::string("db.driver.feature.") + T::getName()].setPointerValue(&feature);
		}
		/**
		 * @brief Registers a new driver feature without implementation.
		 *
		 * This method is used for setting a flag that the driver supports some feature but the feature doesn't need physical implementation.
		 *
		 * @param feature Name of the supported feature
		 */
		inline void setSupportsFeature(const std::string& feature) {
			(*this)[std::string("db.driver.feature.") + feature].setBoolValue(true);
		}

		/**
		 * @brief Checks if the driver supports a feature identified by the name
		 *
		 * @returns true - supports, false - doesn't support feature identified by the name
		 *
		 * @param feature The name of the feature
		 */
		inline bool supportsFeature(const std::string& feature) {
			return hasMeta(std::string("db.driver.feature.") + feature);
		}

		/**
		 * @brief Checks if the driver supports a feature identified by the template parameter
		 *
		 * @returns true - supports, false - doesn't support the feature
		 *
		 */
		template <typename T>
		inline bool supportsFeature() {
			return hasMeta(std::string("db.driver.feature.") + T::getName());
		}

		/**
		 * @brief Returns implementation of the feature T
		 */
		template <typename T>
		T& getFeature() const {
			return *(T*) getMeta(std::string("db.driver.feature.") + T::getName()).getPointerValue();
		}

		//inline DbFeature

	protected:
		// This constructor registers the driver at the DriverManager
		Driver(const std::string& driverName);
		virtual ~Driver();

		// extensions
		inline void registerExtension(const std::string& name, DriverExtension& extension) {
			(*this)[std::string("db.driver.extension.") + name].setPointerValue(&extension);
		}
		inline bool hasExtension(const std::string& extension) {
			return hasMeta(std::string("db.driver.extension.") + extension);
		}
		DriverExtension& getExtension(const std::string& extension) const {
			return *(DriverExtension*) getMeta(std::string("db.driver.extension.") + extension).getPointerValue();
		}


		virtual DatabaseImpl* getDatabaseImpl(const std::string& dbName) = 0;
		virtual void returnDatabaseImpl(DatabaseImpl* dbImpl) = 0;
		virtual void doInit() = 0;
		virtual void doShutdown() = 0;

	private:
		std::string name;

		typedef std::set<Database*> DatabaseSet;
		typedef DatabaseSet::iterator DatabaseSetIterator;

		DatabaseSet databases;

		void addDatabase(Database* db);

		void removeDatabase(Database* db);

		void disconnectDatabases();

		Mutex mutex;

};

}

#endif /*DRIVER_H_*/
