/**
 * @file oracle10gConnectionImpl.cpp
 *
 * @author Josef Troch Josef Troch (<a href="mailto:josef.troch@email.cz">josef.troch@email.cz</a>) 19. - 23.2.2003
 * @author Petr Cermak
 */


#include "oracle10gConnectionImpl.h"
#include "iopccommon/trace.h"
#include <string>
#include "oracle10gCursorImpl.h"

using namespace std;

namespace iopc {


Oracle10gConnectionImpl::Oracle10gConnectionImpl(
	OCIEnv* ociEnv,
	const string& userName,
	const string& password,
	const string& dbName,
	bool autoCommit
	)
{
	IOPC_FUNC_INVOKED;
	ociEnvironment = ociEnv;
	this->userName = userName;
	this->password = password;
	this->dbName = dbName;
	this->autoCommit = autoCommit;
}

void Oracle10gConnectionImpl::open()
{
	IOPC_FUNC_INVOKED;
	sword ret; //OCI return value

	ret = OCIHandleAlloc(ociEnvironment, (dvoid **)&errorHandle, OCI_HTYPE_ERROR, (size_t) 0, NULL);

	if (!OCI_OK(ret)) {
		IOPC_LOG_THROW(IopcDbException(getOCIMsg(ret, "OCIHandleAlloc", true)));
	}

	//create service context handle ("connection" handle)
	ret = OCIHandleAlloc(ociEnvironment, (dvoid **)&serviceContext, OCI_HTYPE_SVCCTX, (size_t) 0, NULL);
	if (!OCI_OK(ret)) {
		const string errMsg = getOCIMsg(ret, "OCIHandleAlloc");
		OCIHandleFree(errorHandle, OCI_HTYPE_ERROR);
		IOPC_LOG_THROW(IopcDbException(errMsg));
	}
	IOPC_DEBUGLOG("OCIHandleAlloc OK");

	//log into DB
	ret = OCILogon(
  		ociEnvironment,
  		errorHandle,
  		&serviceContext,
  		(OraText *) userName.c_str(),
  		userName.size(),
  		(OraText *) password.c_str(),
  		password.size(),
  		(OraText *) dbName.c_str(),
  		dbName.size());

	IOPC_DEBUGLOG("OCILogon OK");

	if (OCI_ERR(ret)) {
		const string errMsg = getOCIMsg(ret, "OCILogon");
		OCIHandleFree(serviceContext, OCI_HTYPE_SVCCTX);
		OCIHandleFree(errorHandle, OCI_HTYPE_ERROR);
		IOPC_LOG_THROW(IopcDbException(errMsg));
	} else if (OCI_MSG(ret)) {
		LogWriter::logInfo(getOCIMsg(ret, "OCILogon"));
	}

	    //run after connect handler
	    /*TODO if (runAfterConnectHandler && afterConnectHandler != NULL)
	    {
	      runAfterConnectHandler = afterConnectHandler(connection);
	    }*/

// TODO _DefaultCursor = new Oracle10gCursor(this);

	IOPC_FUNC_ENDED;
}

void Oracle10gConnectionImpl::close() {
	IOPC_FUNC_INVOKED;

	sword ret = OCILogoff(serviceContext, errorHandle);
	if (OCI_ERR(ret)) {
		IOPC_LOG_THROW(IopcDbException(getOCIMsg(ret, "OCILogoff")));
	} else if (OCI_MSG(ret)) {
		LogWriter::logInfo(getOCIMsg(ret, "OCILogoff"));
	}
	OCIHandleFree(serviceContext, OCI_HTYPE_SVCCTX);
	OCIHandleFree(errorHandle, OCI_HTYPE_ERROR);
	errorHandle = NULL;
	serviceContext = NULL;
}


Oracle10gConnectionImpl::~Oracle10gConnectionImpl()
{
	IOPC_FUNC_INVOKED;
	/*if (isOpened()) {
		IopcLogWriter.logWarning("Deleting connected Oracle10gConnectionImpl. Disconnecting...");
		try {
			disconnect(); // calls doCommit() and doClose()
		} catch (const IopcException& e) {
			IopcLogWriter.logError(string("Error disconnecting db session: ") + e.what());
		}
	}
	if (getCursors().size() > 0) {
		IopcLogWriter.logWarning("Deleting Oracle10gCursors with cursors!");
		closeAllCursors();
	}*/
	IOPC_FUNC_ENDED;
}

//------------------------------------------------------------------------------------

void Oracle10gConnectionImpl::commit()
{
	IOPC_FUNC_INVOKED;
	sword ret = OCITransCommit(serviceContext, errorHandle, OCI_DEFAULT);
	if (OCI_ERR(ret)) {
		IOPC_LOG_THROW(IopcDbException(getOCIMsg(ret, "OCITransCommit")));
	} else if (OCI_MSG(ret)) {
		LogWriter::logInfo(getOCIMsg(ret, "OCITransCommit"));
	}
}

void Oracle10gConnectionImpl::rollback()
{
	IOPC_FUNC_INVOKED;
	sword ret = OCITransRollback(serviceContext, errorHandle, OCI_DEFAULT);
	if (OCI_ERR(ret)) {
		IOPC_LOG_THROW(IopcDbException(getOCIMsg(ret, "OCITransRollback")));
	} else if (OCI_MSG(ret)) {
		LogWriter::logInfo(getOCIMsg(ret, "OCITransRollback"));
	}
}

void Oracle10gConnectionImpl::savepoint(const string& name)
{
	IOPC_FUNC_INVOKED;
	string stmt = "SAVEPOINT " + name;
	sqlNonQuery(stmt); //executes "SAVEPOINT name"
}


void Oracle10gConnectionImpl::rollbackToSavepoint(const string& name)
{
	IOPC_FUNC_INVOKED;
	string stmt = "ROLLBACK TO SAVEPOINT " + name;
	sqlNonQuery(stmt); //executes "ROLLBACK TO SAVEPOINT name"
}

void Oracle10gConnectionImpl::setAutoCommit(bool ac)
{
	autoCommit = ac;
}

CursorImpl* Oracle10gConnectionImpl::getCursor(const std::string& sqlCommand)
{
	IOPC_FUNC_INVOKED;
	return new Oracle10gCursorImpl(ociEnvironment, errorHandle, serviceContext, autoCommit, sqlCommand);
}

void Oracle10gConnectionImpl::returnCursor(CursorImpl* cur)
{
	IOPC_FUNC_INVOKED;
	delete cur;
}

void Oracle10gConnectionImpl::sqlNonQuery(const std::string& sqlCommand)
{
	CursorImpl* cursor = getCursor(sqlCommand);
	cursor->open();
	cursor->prepare();
	cursor->execute();
	returnCursor(cursor);
}

/* TODO Cursor * Oracle10gConnectionImpl::_GetNewCursor()
{
	logDebugMessage("Oracle10gConnectionImpl::_GetNewCursor() invoked");

	Cursor * cursor = new Oracle10gCursor(this);
	if (cursor == NULL) throw ObjLibException_NoMemory();

	return cursor;
}
*/

} // namespace iopc
