#ifndef CACHEPTR_H_
#define CACHEPTR_H_
/**
 * @author Jan Hadrava 2004
 */

#include "cache/cacheLock.h"
#include "persIdentification.h"
#include "db/cachedConnection.h"
#include "iopccommon/countedPtr.h"

namespace iopc {

/** 
 * Template of cache pointer classes.
 * 
 * Valid cache pointer references object locked in cache (for persistent objects).
 */
template<typename T> class CachePtr{
	
public:
	/** Ctor creating invalid cache pointer */
	CachePtr();
	
	/**
	 * Ctor for creating cache pointer to a persistent object
	 */
	CachePtr(CachedConnection& conn, const PersIdentification& persId);

	/**
	 * Ctor for creating cache pointer to a transient object owned by the DbPtr
	 */
	CachePtr(CountedPtr<T> objectPtr);
	
	/** Ctor creating cache pointer to persistent object */
	CachePtr(CacheLock& cLock);
	
	/** Copy ctor */
	CachePtr(const CachePtr<T>& cptr);
	
	~CachePtr();
	
	/** Copy operator */
	CachePtr<T> &operator=(const CachePtr<T>& src);
	
	/** Invalidates cache pointer */
	void release();
	
	/** Inspects cache pointer validity */
	inline bool isValid() const;
	
	/** Inspects cache pointer validity */
	inline operator bool() const;

	/**
	 * Dereferences cache pointer
	 * 
	 * @return Memory pointer to referenced object (NULL for invalid cache pointer).
	 */
	inline CountedPtr<T> operator*() const;
	
	/**
	 * Dereferences cache pointer
	 * 
	 * @return Memory pointer to referenced object (NULL for invalid cache pointer).
	 */
	inline CountedPtr<T> operator->() const;
	
	/**
	 * Dereferences cache pointer
	 *
	 * @return Memory pointer to referenced object (NULL for invalid cache pointer).
	 */
	inline CountedPtr<T> getObjectPtr() const;
	
protected:
	/** Memory pointer, for both persistent and transient instances. */
	CountedPtr<T> objectPtr;
	
	/** Memory pointer for transient instances. */
	//CountedPtr<T> countedObjectPtr;
	
	/** Cache item manipulator (NULL for transient instances). */
	CacheLock* cLock;	
};

template<typename T> CachePtr<T>::CachePtr()
{
	cLock = NULL;
	objectPtr = NULL;
	//countedObjectPtr = NULL;
}

template<typename T> CachePtr<T>::CachePtr(CachedConnection& conn, const PersIdentification& persId)
{
	cLock = &conn.getCacheLock(persId, false);
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock->getObject()), false);
	//countedObjectPtr = NULL;
}

template<typename T> CachePtr<T>::CachePtr(CountedPtr<T> object)
{
	cLock = NULL;
	//this->objectPtr = NULL;
	this->objectPtr =  object;
	//dynamic_cast<T*>(objectPtr);
}

template<typename T> CachePtr<T>::CachePtr(const CachePtr<T>& src)
{
	objectPtr = src.objectPtr;
	//countedObjectPtr = src.countedObjectPtr;
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(false);
	}
}

template<typename T> CachePtr<T>::CachePtr(CacheLock& cLock)
{
	this->cLock = &cLock;
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock.getObject()), false);
	//countedObjectPtr = NULL;
}

template<typename T> CachePtr<T>::~CachePtr()
{
	IOPC_FUNC_INVOKED;
	release();
}

template<typename T> CachePtr<T>& CachePtr<T>::operator=(const CachePtr<T>& src)
{
	if (this == &src)
		return *this;
	objectPtr = src.objectPtr;
	//countedObjectPtr = src.countedObjectPtr;
	if (cLock != NULL) {
		cLock->rmLock();
	}
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(false);
		IOPC_ASSERT(objectPtr == (T *)&cLock->getObject());
	}
	return *this;
}

template<typename T> void CachePtr<T>::release()
{
	IOPC_FUNC_INVOKED;
	if (cLock != NULL) {
		cLock->rmLock();
		cLock = NULL;
	}
	objectPtr = CountedPtr<T>();
	//countedObjectPtr = NULL;
	IOPC_FUNC_ENDED;
}

template<typename T> inline bool CachePtr<T>::isValid() const
{
	return objectPtr != NULL;// || countedObjectPtr != NULL;
}

template<typename T> inline CachePtr<T>::operator bool() const
{
	return isValid();
}

template<typename T> CountedPtr<T> CachePtr<T>::operator*() const
{
	return getObjectPtr();
}

template<typename T> CountedPtr<T> CachePtr<T>::operator->() const
{
	return getObjectPtr();
}

template<typename T> CountedPtr<T> CachePtr<T>::getObjectPtr() const
{
	return objectPtr;
}

} /*namespace*/



#endif /*CACHEPTR_H_*/
