#ifndef CACHEDCONNECTION_H_
#define CACHEDCONNECTION_H_

#include "iopcdb/connection.h"
#include "../cache/cache.h"
#include "../cache/cacheLock.h"
#include "../persIdentification.h"
#include "iopccommon/rwLock.h"
#include "cachedDatabase.h"

namespace iopc {


/**
 * 
 * NOTE: Doesn't allow changing the associated cache any more.
 */

class CachedConnection : public Connection
{
private:
	Connection* inner;
	Cache* cache;
	CachedDatabase& database;
	RWLock cacheLock;
	
public:

	CachedConnection(Connection* connection, Cache* cache, CachedDatabase& database);
	
	void stopCache();
	void restartCache();
	
	/**
	 * Deletes the inner instance of Connection and the cache
	 */
	virtual ~CachedConnection();
	
	// Connection interface
	
	/** Passes an SQL command to DB. */
	virtual Cursor* sql(const std::string& name);
	
	/** bypasses cache */
	virtual Cursor* sqlNoCache(const std::string& name);
	
	//virtual Cursor* getCursor(const std::string& name);
	
	virtual void returnCursor(Cursor* cur);
	
	virtual void closeAllCursors();
	
	virtual void sqlNonQuery(const std::string& name);
	
	/** bypasses cache */
	virtual void sqlNonQueryNoCache(const std::string& name);
	
	/** Commits current transaction. Starts new one. */
	virtual void commit();
	/** Rollbacks current transaction.*/
	virtual void rollback();
	/** Creates a savepoint. */
	virtual void savepoint(const std::string& name);
	/** Rollbacks to savepoint. */
	virtual void rollbackToSavepoint(const std::string& name);
	/** Opens connection */
	virtual void open();
	/** Closes connection */
	virtual void close();
	/** Commits last transaction and disconnects from DB. */
	virtual void disconnect();
	/** Rollbacks current trans
	 * action and disconnects from DB. */
	virtual void abort();
	/** Turns on/off autocommit. */
	virtual void setAutoCommit(bool ac);
	/** Returns state of this connection */
	virtual bool isOpened();
	/** Passes an SQL command to DB. */
	virtual Connection& operator << (
	    const std::string& command
	);
	
	//Tests, if the Database connections are the same
	virtual bool operator == (const Connection& dbConn) const;
	//Tests, if the Database connections are not the same
	virtual bool operator != (const Connection& dbConn) const;
	
	inline Database& getDatabase() { return database; }
	
	inline Driver& getDriver() { return database.getDriver(); }
	
	virtual inline bool isAutoCommit() { return inner->isAutoCommit(); }
	
	virtual inline void updateDirty() { cache->updateDirty(); }

	virtual inline void removeAll() { cache->removeAll(); }
	
	inline ConnectionImpl* getImplementation() { return inner->getImplementation(); }
	
	virtual Connection* getBasicConnection();
	virtual Connection* getInner();
	
	CacheLock& makePersistent(const PersIdentification& persId, DatabaseObject& objectPtr);
	
	CacheLock& getCacheLock(const PersIdentification& persId, bool rdonly = false);
	
	void dbDelete(const PersIdentification& persId);
	
	void cacheDelete(const PersIdentification& persId);

	virtual inline Connection* getInnerConnection() { return inner; }
	
	inline Cache& getCache() { return *cache; }

/*
	// generates new OID and Serial Number for new object
	virtual bool doGetNextOID(long int& nxtoid, long int& nxtsn) = 0;
	// generates new Serial Number for object
	virtual bool doGetNextSN(long int curoid, long int& nxtsn) = 0;*/
	
	
	/** Temporarily blocks cache usage. */
	//void StopCache();
	/** Unblocks cache usage. */
	//void RestartCache();
	
};

}

#endif /*CACHEDCONNECTION_H_*/
